#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import time
import unittest

from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.core.unit import Unit, UnitState
from fabric_cf.actor.core.kernel.reservation_client import ClientReservationFactory
from fabric_cf.actor.core.kernel.slice import SliceFactory
from fabric_cf.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.resource_type import ResourceType
from fabric_cf.actor.test.base_test_case import BaseTestCase


class UnitTest(BaseTestCase, unittest.TestCase):
    from fabric_cf.actor.core.container.globals import Globals
    Globals.config_file = "./config/config.test.yaml"
    Constants.SUPERBLOCK_LOCATION = './state_recovery.lock'

    from fabric_cf.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def make_actor_database(self) -> SubstrateActorDatabase:
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        return GlobalsSingleton.get().get_container().get_actor().get_plugin().get_database()

    def test_unit(self):
        rid = ID()
        u1 = Unit(rid=rid)
        self.assertIsNotNone(u1.get_id())
        self.assertEqual(UnitState.DEFAULT, u1.get_state())
        self.assertIsNone(u1.get_property(name="foo"))
        self.assertIsNone(u1.get_parent_id())
        self.assertIsNotNone(u1.get_reservation_id())
        self.assertIsNone(u1.get_slice_id())
        self.assertIsNone(u1.get_actor_id())

        self.assertEqual(0, u1.get_sequence())
        u1.increment_sequence()
        self.assertEqual(1, u1.get_sequence())
        u1.decrement_sequence()
        self.assertEqual(0, u1.get_sequence())

        db = self.make_actor_database()

        slice_id = ID()
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        actor_id = GlobalsSingleton.get().get_container().get_actor().get_guid()

        slice_obj = SliceFactory.create(slice_id=slice_id, name="test_slice")
        db.add_slice(slice_object=slice_obj)

        reservation = ClientReservationFactory.create(rid=rid, slice_object=slice_obj)
        u1.set_actor_id(actor_id=actor_id)
        u1.set_reservation(reservation=reservation)
        u1.set_slice_id(slice_id=slice_id)

        db.add_reservation(reservation=reservation)

        u1.start_prime()
        self.assertEqual(UnitState.PRIMING, u1.get_state())
        u1.set_property(name="foo", value="bar")
        u1.increment_sequence()
        u1.increment_sequence()
        resource_type = ResourceType(resource_type="1")
        u1.set_resource_type(rtype=resource_type)
        self.assertEqual(2, u1.get_sequence())

        db.add_unit(u=u1)

        self.assertIsNotNone(db.get_unit(uid=rid))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import ipaddress
import random
import traceback
from ipaddress import IPv6Network, IPv4Network
from typing import List, Tuple, Union

from fabric_cf.actor.fim.fim_helper import FimHelper
from fim.slivers.capacities_labels import Labels
from fim.slivers.gateway import Gateway
from fim.slivers.interface_info import InterfaceSliver, InterfaceType
from fim.slivers.network_node import NodeSliver
from fim.slivers.network_service import NetworkServiceSliver, NSLayer, ServiceType

from fabric_cf.actor.core.apis.abc_reservation_mixin import ABCReservationMixin
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import BrokerException, ExceptionErrorCode
from fabric_cf.actor.core.kernel.reservation_states import ReservationOperation
from fabric_cf.actor.core.policy.inventory_for_type import InventoryForType
from fabric_cf.actor.core.util.id import ID


class NetworkServiceInventory(InventoryForType):
    @staticmethod
    def __extract_vlan_range(*, labels: Labels) -> List[int] or None:
        vlan_range = None
        if labels is None:
            return vlan_range
        if labels.vlan_range is not None:
            if isinstance(labels.vlan_range, list):
                vlan_range = []
                for v_r in labels.vlan_range:
                    vlans = v_r.split("-")
                    for x in list(range(int(vlans[0]), int(vlans[1]) + 1)):
                        if x not in vlan_range:
                            vlan_range.append(x)
            else:
                vlans = labels.vlan_range.split("-")
                vlan_range = list(range(int(vlans[0]), int(vlans[1]) + 1))
        elif labels.vlan is not None:
            vlan_range = [int(labels.vlan)]
        return vlan_range

    def __exclude_allocated_vlans(self, *, rid: ID, available_vlan_range: List[int], bqm_ifs: InterfaceSliver,
                                  existing_reservations: List[ABCReservationMixin]) -> List[int]:
        # Exclude the already allocated VLANs and subnets
        if existing_reservations is None:
            return available_vlan_range

        for reservation in existing_reservations:
            if rid == reservation.get_reservation_id():
                continue

            # For Active or Ticketed or Ticketing reservations; reduce the counts from available
            allocated_sliver = self.get_allocated_sliver(reservation=reservation)

            self.logger.debug(
                f"Existing res# {reservation.get_reservation_id()} state:{reservation.get_state()} "
                f"allocated: {allocated_sliver}")

            if allocated_sliver is None:
                continue

            if allocated_sliver.interface_info is None:
                continue

            for allocated_ifs in allocated_sliver.interface_info.interfaces.values():
                # Ignore the Interface Slivers not on the same port
                if allocated_ifs.get_node_map()[1] != bqm_ifs.node_id:
                    continue

                self.logger.debug(f"Excluding already allocated VLAN: "
                                  f"{allocated_ifs.label_allocations.vlan} to "
                                  f"res# {reservation.get_reservation_id()}")

                # Exclude VLANs on the allocated on the same port
                if allocated_ifs.label_allocations.vlan is not None:
                    allocated_vlan = int(allocated_ifs.label_allocations.vlan)

                    if allocated_vlan in available_vlan_range:
                        available_vlan_range.remove(allocated_vlan)

        if available_vlan_range is None or len(available_vlan_range) == 0:
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"No VLANs available!")
        return available_vlan_range

    def allocate_ifs(self, *, rid: ID, requested_ns: NetworkServiceSliver, requested_ifs: InterfaceSliver,
                     owner_ns: NetworkServiceSliver, bqm_ifs: InterfaceSliver,
                     existing_reservations: List[ABCReservationMixin],
                     operation: ReservationOperation = ReservationOperation.Create) -> InterfaceSliver:
        """
        Allocate Interface Sliver
        - For L2 services, validate the VLAN tag specified is within the allowed range
        - For L3 services,
            - grab the VLAN from BQM Site specific NetworkService
            - exclude the VLAN already assigned to other Interface Sliver on the same port
            - allocate the first available VLAN to the Interface Sliver
        :param rid: Reservation ID
        :param requested_ns: Requested NetworkService
        :param requested_ifs: Requested Interface Sliver
        :param owner_ns: BQM NetworkService identified to serve the InterfaceSliver
        :param bqm_ifs: BQM InterfaceSliver identified to serve the InterfaceSliver
        :param existing_reservations: Existing Reservations which also are served by the owner switch
        :param operation: Extend/Create/Modify Operation
        :return Interface Sliver updated with the allocated VLAN tag for FABNetv4 and FABNetv6 services
        :raises Exception if vlan tag range is not in the valid range for L2 services
        """
        requested_vlan = None
        if requested_ifs.labels and requested_ifs.labels.vlan:
            requested_vlan = int(requested_ifs.labels.vlan)

        if requested_ns.get_layer() == NSLayer.L2:
            # Validate the requested VLAN is in range specified on MPLS Network Service in BQM
            # Only do this for Non FacilityPorts
            if bqm_ifs.get_type() != InterfaceType.FacilityPort:
                if requested_vlan is None:
                    return requested_ifs

                if owner_ns.get_label_delegations() is None:
                    if not (1 <= requested_vlan <= 4095):
                        raise BrokerException(
                            error_code=ExceptionErrorCode.FAILURE,
                            msg=f"Vlan for L2 service {requested_vlan} is outside the allowed range 1-4095"
                        )
                    return requested_ifs

                delegation_id, delegated_label = FimHelper.get_delegations(delegations=owner_ns.get_label_delegations())
                vlan_range = self.__extract_vlan_range(labels=delegated_label)

                if vlan_range and requested_vlan not in vlan_range:
                    raise BrokerException(
                        error_code=ExceptionErrorCode.FAILURE,
                        msg=f"Vlan for L2 service {requested_vlan} is outside the available range {vlan_range}"
                    )

            vlan_range = self.__extract_vlan_range(labels=bqm_ifs.labels)
            if vlan_range:
                vlan_range = self.__exclude_allocated_vlans(rid=rid, available_vlan_range=vlan_range,
                                                            bqm_ifs=bqm_ifs,
                                                            existing_reservations=existing_reservations)

                if operation == ReservationOperation.Extend:
                    if requested_vlan and requested_vlan not in vlan_range:
                        raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                              msg=f"Renew failed: VLAN {requested_vlan} for Interface : "
                                                  f"{requested_ifs.get_name()/bqm_ifs.node_id} already in "
                                                  f"use by another reservation")
                    return requested_ifs

                if requested_vlan is None:
                    requested_ifs.labels.vlan = str(random.choice(vlan_range))
                    return requested_ifs

                if requested_vlan not in vlan_range:
                    raise BrokerException(
                        error_code=ExceptionErrorCode.FAILURE,
                        msg=f"Vlan for L2 service {requested_vlan} is outside the available range {vlan_range}"
                    )
        else:
            if bqm_ifs.get_type() != InterfaceType.FacilityPort:
                delegation_id, delegated_label = FimHelper.get_delegations(
                    delegations=owner_ns.get_label_delegations())
                vlan_range = self.__extract_vlan_range(labels=delegated_label)
            else:
                vlan_range = self.__extract_vlan_range(labels=bqm_ifs.labels)

            if vlan_range:
                vlan_range = self.__exclude_allocated_vlans(rid=rid, available_vlan_range=vlan_range,
                                                            bqm_ifs=bqm_ifs,
                                                            existing_reservations=existing_reservations)

                if operation == ReservationOperation.Extend:
                    if requested_vlan and requested_vlan not in vlan_range:
                        raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                              msg=f"Renew failed: VLAN {requested_vlan} for Interface : "
                                                  f"{requested_ifs.get_name()}/{bqm_ifs.node_id} already in "
                                                  f"use by another reservation")
                    return requested_ifs

                if bqm_ifs.get_type() != InterfaceType.FacilityPort:
                    requested_ifs.labels.vlan = str(random.choice(vlan_range))
                    requested_ifs.label_allocations = Labels(vlan=requested_ifs.labels.vlan)
                else:
                    if not requested_ifs.labels:
                        return requested_ifs

                    if requested_ifs.labels.vlan is None:
                        requested_ifs.labels.vlan = str(random.choice(vlan_range))

                    if int(requested_ifs.labels.vlan) not in vlan_range:
                        raise BrokerException(
                            error_code=ExceptionErrorCode.FAILURE,
                            msg=f"Vlan for L3 service {requested_ifs.labels.vlan} is outside the "
                                f"available range {vlan_range}"
                        )

        return requested_ifs

    def __allocate_ip_address_to_ifs(self, *, requested_ns: NetworkServiceSliver) -> NetworkServiceSliver:
        """
        Allocate IP addresses to the interfaces of the requested network service sliver.

        :param requested_ns: The requested network service sliver.
        :return: The updated network service sliver with allocated IP addresses.
        """
        if requested_ns.gateway is None:
            return requested_ns

        if requested_ns.get_type() in Constants.L3_FABNET_EXT_SERVICES:
            return requested_ns

        if requested_ns.get_type() == ServiceType.FABNetv4:
            start_ip_str = requested_ns.gateway.lab.ipv4
        else:
            start_ip_str = requested_ns.gateway.lab.ipv6

        start_ip = ipaddress.ip_address(start_ip_str)
        start_ip += 1

        for ifs in requested_ns.interface_info.interfaces.values():
            if requested_ns.get_type() == ServiceType.FABNetv4:
                ifs.labels.ipv4 = str(start_ip)
                ifs.label_allocations.ipv4 = str(start_ip)

            elif requested_ns.get_type() == ServiceType.FABNetv6:
                ifs.labels.ipv6 = str(start_ip)
                ifs.label_allocations.ipv6 = str(start_ip)

            self.logger.info("Allocated IP address to interface %s", ifs)
            start_ip += 1

        return requested_ns

    def allocate_vnic(self, *, rid: ID, requested_ns: NetworkServiceSliver, owner_ns: NetworkServiceSliver,
                      existing_reservations: List[ABCReservationMixin]) -> NetworkServiceSliver:
        """
        Allocate Network Service Sliver (Only for L2Bridge Service for OpenStackvNIC)
        :param rid: Reservation ID
        :param requested_ns: Requested NetworkService
        :param owner_ns: BQM Network Service Sliver
        :param existing_reservations: Existing Reservations which also are served by the owner switch
        :return NetworkService updated with the allocated vlan
        """
        try:
            if requested_ns.get_type() != ServiceType.L2Bridge:
                return requested_ns

            vlans_range = self.__extract_vlan_range(labels=owner_ns.labels)

            # Exclude the already allocated VLANs
            for reservation in existing_reservations:
                if rid == reservation.get_reservation_id():
                    continue

                # For Active or Ticketed or Ticketing reservations; reduce the counts from available
                allocated_sliver = self.get_allocated_sliver(reservation=reservation)

                self.logger.debug(f"Existing res# {reservation.get_reservation_id()} "
                                  f"allocated: {allocated_sliver}")

                if allocated_sliver is None:
                    continue

                if allocated_sliver.get_type() != requested_ns.get_type():
                    continue

                if allocated_sliver.label_allocations is not None and allocated_sliver.label_allocations.vlan is not None:
                    vlans_range.remove(int(allocated_sliver.label_allocations.vlan))

            if requested_ns.label_allocations is None:
                requested_ns.label_allocations = Labels()
            #requested_ns.label_allocations.vlan = str(random.choice(vlans_range))
            requested_ns.label_allocations.vlan = str(vlans_range[0])
        except Exception as e:
            self.logger.error(f"Error in allocate_vNIC: {e}")
            self.logger.error(traceback.format_exc())
            raise BrokerException(msg=f"Allocation failure for Openstack VNIC: {e}")
        return requested_ns

    def allocate(self, *, rid: ID, requested_ns: NetworkServiceSliver, owner_ns: NetworkServiceSliver,
                 existing_reservations: List[ABCReservationMixin]) -> NetworkServiceSliver:
        """
        Allocate Network Service Sliver (Only for L3 Service)
            - grab the /17 or /48 from BQM Site specific NetworkService
            - divide it into /24 or /64 subnets
            - exclude the 1st subnet (reserved for control plane)
            - exclude the subnets already assigned to other V4/V6 NetworkService on the same owner switch
            - allocate the first available subnet to the NetworkService
        :param rid: Reservation ID
        :param requested_ns: Requested NetworkService
        :param owner_ns: BQM Network Service identified to serve the NetworkService
        :param existing_reservations: Existing Reservations which also are served by the owner switch
        :return: NetworkService updated with the allocated subnet for FABNetv4 and FABNetv6 services
        """
        try:
            if requested_ns.get_type() not in Constants.L3_SERVICES:
                return requested_ns

            # Grab Label Delegations
            delegation_id, delegated_label = FimHelper.get_delegations(delegations=owner_ns.get_label_delegations())

            # HACK to use FabNetv6 for FabNetv6Ext as both have the same range
            requested_ns_type = requested_ns.get_type()
            if requested_ns_type == ServiceType.FABNetv6Ext:
                requested_ns_type = ServiceType.FABNetv6

            # Handle L3VPN type specifically
            if requested_ns_type == ServiceType.L3VPN:
                if requested_ns.labels is not None:
                    requested_ns.labels = Labels.update(requested_ns.labels, asn=delegated_label.asn)
                else:
                    requested_ns.labels = Labels(asn=delegated_label.asn)
                return requested_ns

            ip_network, subnet_list = self._generate_subnet_list(owner_ns=owner_ns, delegated_label=delegated_label)

            # Exclude the already allocated subnets
            subnet_list = self._exclude_allocated_subnets(subnet_list=subnet_list, requested_ns_type=requested_ns_type,
                                                          rid=rid, existing_reservations=existing_reservations)

            # Extend Case
            if requested_ns.get_node_map():
                self._can_extend(subnet_list=subnet_list, requested_ns=requested_ns)
                return requested_ns

            gateway_labels = self._assign_gateway_labels(ip_network=ip_network, subnet_list=subnet_list,
                                                         requested_ns=requested_ns)

            self.logger.debug(f"Gateway Labels: {gateway_labels}")

            requested_ns.gateway = Gateway(lab=gateway_labels)

            # Allocate the IP Addresses for the requested NS
            requested_ns = self.__allocate_ip_address_to_ifs(requested_ns=requested_ns)
        except BrokerException as e:
            raise e
        except Exception as e:
            self.logger.error(f"Error in allocate_gateway_for_ns: {e}")
            self.logger.error(traceback.format_exc())
            raise BrokerException(msg=f"Allocation failure for Requested Network Service: {e}")

        return requested_ns

    def _generate_subnet_list(self, *, owner_ns: NetworkServiceSliver,
                              delegated_label: Labels) -> Tuple[Union[IPv4Network, IPv6Network], List]:
        """
        Generate the list of subnets based on the owner network service type.

        :param owner_ns: The NetworkServiceSliver representing the owner network service.
        :param delegated_label: The Labels object containing the delegated subnet information.
        :return: A tuple containing the IP network and the list of generated subnets.
        """
        subnet_list = None
        ip_network = None
        if owner_ns.get_type() in Constants.L3_FABNETv6_SERVICES:
            ip_network = IPv6Network(delegated_label.ipv6_subnet)
            subnet_list = list(ip_network.subnets(new_prefix=64))
            # Exclude the 1st subnet as it is reserved for control plane
            subnet_list.pop(0)
            # Exclude the last subnet for FABRIC STAR Bastion Host Allocation
            subnet_list.pop(-1)

        elif owner_ns.get_type() in [ServiceType.FABNetv4, ServiceType.FABNetv4Ext]:
            ip_network = IPv4Network(delegated_label.ipv4_subnet)
            if owner_ns.get_type() == ServiceType.FABNetv4:
                subnet_list = list(ip_network.subnets(new_prefix=24))
                subnet_list.pop(0)

            elif owner_ns.get_type() == ServiceType.FABNetv4Ext:
                subnet_list = list(ip_network.hosts())

        self.logger.debug(f"Available Subnets: {subnet_list}")

        return ip_network, subnet_list

    def _exclude_allocated_subnets(self, *, subnet_list: List, requested_ns_type: str, rid: ID,
                                   existing_reservations: List[ABCReservationMixin]) -> List:
        """
        Exclude the subnets that are already allocated.

        :param subnet_list: A list of available subnets to be allocated.
        :param requested_ns_type: The type of the requested network service.
        :param rid: The reservation ID of the current request.
        :param existing_reservations: A list of existing reservations that may contain allocated subnets.
        :return: A list of subnets excluding those that have already been allocated.
        """
        for reservation in existing_reservations:
            if rid == reservation.get_reservation_id():
                continue

            allocated_sliver = self.get_allocated_sliver(reservation)
            if allocated_sliver is None:
                continue

            # HACK to use FabNetv6 for FabNetv6Ext as both have the same range
            # Needs to be removed if FabNetv6/FabNetv6Ext are configured with different ranges
            allocated_sliver_type = allocated_sliver.get_type()
            if allocated_sliver_type == ServiceType.FABNetv6Ext:
                allocated_sliver_type = ServiceType.FABNetv6
            # HACK End

            if allocated_sliver_type != requested_ns_type:
                continue

            if allocated_sliver.get_type() == ServiceType.FABNetv4:
                subnet_to_remove = IPv4Network(allocated_sliver.get_gateway().subnet)
                self.logger.debug(
                    f"Excluding already allocated IP4Subnet: "
                    f"{allocated_sliver.get_gateway().subnet}"
                    f" to res# {reservation.get_reservation_id()}")
                if subnet_to_remove in subnet_list:
                    subnet_list.remove(subnet_to_remove)

            elif allocated_sliver.get_type() == ServiceType.FABNetv4Ext:
                if allocated_sliver.labels is not None and allocated_sliver.labels.ipv4 is not None:
                    for x in allocated_sliver.labels.ipv4:
                        subnet_to_remove = ipaddress.IPv4Address(x)
                        self.logger.debug(
                            f"Excluding already allocated IP4: "
                            f"{x}"
                            f" to res# {reservation.get_reservation_id()}")
                        if subnet_to_remove in subnet_list:
                            subnet_list.remove(subnet_to_remove)

            elif allocated_sliver.get_type() in Constants.L3_FABNETv6_SERVICES:
                subnet_to_remove = IPv6Network(allocated_sliver.get_gateway().subnet)
                self.logger.debug(
                    f"Excluding already allocated IP6Subnet: "
                    f"{allocated_sliver.get_gateway().subnet}"
                    f" to res# {reservation.get_reservation_id()}")

                if subnet_to_remove in subnet_list:
                    subnet_list.remove(subnet_to_remove)

            self.logger.debug(f"Excluding already allocated subnet for reservation {reservation.get_reservation_id()}")

        return subnet_list

    def _assign_gateway_labels(self, *, ip_network: Union[IPv4Network, IPv6Network], subnet_list: List,
                               requested_ns: NetworkServiceSliver) -> Labels:
        """
        Assign gateway labels based on the requested network service type.

        :param ip_network: The IP network from which subnets are derived, either IPv4Network or IPv6Network.
        :param subnet_list: A list of subnets derived from the ip_network.
        :param requested_ns: Network Service sliver.
        :return: Gateway labels populated with the appropriate subnet and IP address.
        """
        if len(subnet_list) == 0:
            raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                  msg=f"No subnets available for {requested_ns.get_site()}")
        gateway_labels = Labels()
        if requested_ns.get_type() == ServiceType.FABNetv4:
            # Allocate the requested network if available else allocate new network
            if requested_ns.gateway and requested_ns.gateway.lab and requested_ns.gateway.lab.ipv4_subnet:
                requested_subnet = IPv4Network(requested_ns.gateway.lab.ipv4_subnet)
                if requested_subnet in subnet_list:
                    gateway_labels.ipv4_subnet = requested_subnet.with_prefixlen
                    gateway_labels.ipv4 = str(next(requested_subnet.hosts()))
                    return gateway_labels

            gateway_labels.ipv4_subnet = subnet_list[0].with_prefixlen
            gateway_labels.ipv4 = str(list(subnet_list[0].hosts())[0])

        elif requested_ns.get_type() == ServiceType.FABNetv4Ext:
            gateway_labels.ipv4_subnet = ip_network.with_prefixlen
            gateway_labels.ipv4 = str(subnet_list[0])

        elif requested_ns.get_type() in Constants.L3_FABNETv6_SERVICES:
            # Allocate the requested network if available else allocate new network
            if requested_ns.gateway and requested_ns.gateway.lab and requested_ns.gateway.lab.ipv6_subnet:
                requested_subnet = IPv6Network(requested_ns.gateway.lab.ipv6_subnet)
                if requested_subnet in subnet_list:
                    gateway_labels.ipv6_subnet = requested_subnet.with_prefixlen
                    gateway_labels.ipv6 = str(next(requested_subnet.hosts()))
                    return gateway_labels

            gateway_labels.ipv6_subnet = subnet_list[0].with_prefixlen
            gateway_labels.ipv6 = str(next(subnet_list[0].hosts()))

        self.logger.debug(f"Allocated Gateway Labels for Network Service: {gateway_labels}")

        return gateway_labels

    def _can_extend(self, *, subnet_list: List, requested_ns: NetworkServiceSliver):
        if requested_ns.get_type() == ServiceType.FABNetv4:
            allocated_subnet = ipaddress.IPv4Network(requested_ns.gateway.subnet)
            if allocated_subnet not in subnet_list:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Renew failed: Subnet {requested_ns.gateway.subnet} for Network Service : "
                                          f"{requested_ns.get_type()} already in use by another reservation")
        elif requested_ns.get_type() == ServiceType.FABNetv4Ext:
            if requested_ns.gateway.gateway in subnet_list:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Renew failed: Subnet {requested_ns.gateway.subnet} for Network Service : "
                                          f"{requested_ns.get_type()} already in use by another reservation")

        elif requested_ns.get_type() in Constants.L3_FABNETv6_SERVICES:
            allocated_subnet = ipaddress.IPv6Network(requested_ns.gateway.subnet)
            if allocated_subnet not in subnet_list:
                raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                      msg=f"Renew failed: Subnet {requested_ns.gateway.subnet} for Network Service : "
                                          f"{requested_ns.get_type()} already in use by another reservation")

    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        pass

    def allocate_peered_ifs(self, *, rid: ID, owner_switch: NodeSliver,
                            requested_ifs: InterfaceSliver, bqm_interface: InterfaceSliver,
                            existing_reservations: List[ABCReservationMixin]) -> InterfaceSliver:
        """
        Update Labels for a Peered Interface
        @param
        """
        ifs_labels = requested_ifs.get_labels()
        if ifs_labels is None:
            ifs_labels = Labels()

        if owner_switch.get_name() == Constants.AL2S:
            delegation_id, delegated_label = FimHelper.get_delegations(delegations=bqm_interface.get_label_delegations())
            local_name = delegated_label.local_name
            device_name = delegated_label.device_name
        else:
            local_name = bqm_interface.get_name()
            device_name = owner_switch.get_name()

        ifs_labels = Labels.update(ifs_labels, local_name=local_name, device_name=device_name)

        if bqm_interface.labels.vlan_range is not None:
            vlan_range = self.__extract_vlan_range(labels=bqm_interface.labels)
            available_vlans = self.__exclude_allocated_vlans(rid=rid, available_vlan_range=vlan_range, bqm_ifs=bqm_interface,
                                                             existing_reservations=existing_reservations)

            # Extend case
            if requested_ifs.get_node_map() and requested_ifs.labels and requested_ifs.labels.vlan:
                if int(requested_ifs.labels.vlan) in available_vlans:
                    raise BrokerException(error_code=ExceptionErrorCode.INSUFFICIENT_RESOURCES,
                                          msg=f"Renew failed: VLAN {requested_ifs.labels.vlan} for Interface : "
                                              f"{requested_ifs.get_name()} already in use by another reservation")

            vlan = str(random.choice(available_vlans))
            #vlan = str(available_vlans[0])
            ifs_labels = Labels.update(ifs_labels, vlan=vlan)

        requested_ifs.labels = ifs_labels
        requested_ifs.label_allocations = Labels.update(lab=ifs_labels)

        return requested_ifs



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class Sliver(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, notice: str=None, sliver_type: str=None, sliver: object=None, lease_start_time: str=None, lease_end_time: str=None, state: str=None, pending_state: str=None, join_state: str=None, graph_node_id: str=None, slice_id: str=None, sliver_id: str=None, owner_user_id: str=None, owner_email: str=None):  # noqa: E501
        """Sliver - a model defined in Swagger

        :param notice: The notice of this Sliver.  # noqa: E501
        :type notice: str
        :param sliver_type: The sliver_type of this Sliver.  # noqa: E501
        :type sliver_type: str
        :param sliver: The sliver of this Sliver.  # noqa: E501
        :type sliver: object
        :param lease_start_time: The lease_start_time of this Sliver.  # noqa: E501
        :type lease_start_time: str
        :param lease_end_time: The lease_end_time of this Sliver.  # noqa: E501
        :type lease_end_time: str
        :param state: The state of this Sliver.  # noqa: E501
        :type state: str
        :param pending_state: The pending_state of this Sliver.  # noqa: E501
        :type pending_state: str
        :param join_state: The join_state of this Sliver.  # noqa: E501
        :type join_state: str
        :param graph_node_id: The graph_node_id of this Sliver.  # noqa: E501
        :type graph_node_id: str
        :param slice_id: The slice_id of this Sliver.  # noqa: E501
        :type slice_id: str
        :param sliver_id: The sliver_id of this Sliver.  # noqa: E501
        :type sliver_id: str
        :param owner_user_id: The owner_user_id of this Sliver.  # noqa: E501
        :type owner_user_id: str
        :param owner_email: The owner_email of this Sliver.  # noqa: E501
        :type owner_email: str
        """
        self.swagger_types = {
            'notice': str,
            'sliver_type': str,
            'sliver': object,
            'lease_start_time': str,
            'lease_end_time': str,
            'state': str,
            'pending_state': str,
            'join_state': str,
            'graph_node_id': str,
            'slice_id': str,
            'sliver_id': str,
            'owner_user_id': str,
            'owner_email': str
        }

        self.attribute_map = {
            'notice': 'notice',
            'sliver_type': 'sliver_type',
            'sliver': 'sliver',
            'lease_start_time': 'lease_start_time',
            'lease_end_time': 'lease_end_time',
            'state': 'state',
            'pending_state': 'pending_state',
            'join_state': 'join_state',
            'graph_node_id': 'graph_node_id',
            'slice_id': 'slice_id',
            'sliver_id': 'sliver_id',
            'owner_user_id': 'owner_user_id',
            'owner_email': 'owner_email'
        }
        self._notice = notice
        self._sliver_type = sliver_type
        self._sliver = sliver
        self._lease_start_time = lease_start_time
        self._lease_end_time = lease_end_time
        self._state = state
        self._pending_state = pending_state
        self._join_state = join_state
        self._graph_node_id = graph_node_id
        self._slice_id = slice_id
        self._sliver_id = sliver_id
        self._owner_user_id = owner_user_id
        self._owner_email = owner_email

    @classmethod
    def from_dict(cls, dikt) -> 'Sliver':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The sliver of this Sliver.  # noqa: E501
        :rtype: Sliver
        """
        return util.deserialize_model(dikt, cls)

    @property
    def notice(self) -> str:
        """Gets the notice of this Sliver.


        :return: The notice of this Sliver.
        :rtype: str
        """
        return self._notice

    @notice.setter
    def notice(self, notice: str):
        """Sets the notice of this Sliver.


        :param notice: The notice of this Sliver.
        :type notice: str
        """

        self._notice = notice

    @property
    def sliver_type(self) -> str:
        """Gets the sliver_type of this Sliver.


        :return: The sliver_type of this Sliver.
        :rtype: str
        """
        return self._sliver_type

    @sliver_type.setter
    def sliver_type(self, sliver_type: str):
        """Sets the sliver_type of this Sliver.


        :param sliver_type: The sliver_type of this Sliver.
        :type sliver_type: str
        """

        self._sliver_type = sliver_type

    @property
    def sliver(self) -> object:
        """Gets the sliver of this Sliver.


        :return: The sliver of this Sliver.
        :rtype: object
        """
        return self._sliver

    @sliver.setter
    def sliver(self, sliver: object):
        """Sets the sliver of this Sliver.


        :param sliver: The sliver of this Sliver.
        :type sliver: object
        """

        self._sliver = sliver

    @property
    def lease_start_time(self) -> str:
        """Gets the lease_start_time of this Sliver.


        :return: The lease_start_time of this Sliver.
        :rtype: str
        """
        return self._lease_start_time

    @lease_start_time.setter
    def lease_start_time(self, lease_start_time: str):
        """Sets the lease_start_time of this Sliver.


        :param lease_start_time: The lease_start_time of this Sliver.
        :type lease_start_time: str
        """

        self._lease_start_time = lease_start_time

    @property
    def lease_end_time(self) -> str:
        """Gets the lease_end_time of this Sliver.


        :return: The lease_end_time of this Sliver.
        :rtype: str
        """
        return self._lease_end_time

    @lease_end_time.setter
    def lease_end_time(self, lease_end_time: str):
        """Sets the lease_end_time of this Sliver.


        :param lease_end_time: The lease_end_time of this Sliver.
        :type lease_end_time: str
        """

        self._lease_end_time = lease_end_time

    @property
    def state(self) -> str:
        """Gets the state of this Sliver.


        :return: The state of this Sliver.
        :rtype: str
        """
        return self._state

    @state.setter
    def state(self, state: str):
        """Sets the state of this Sliver.


        :param state: The state of this Sliver.
        :type state: str
        """

        self._state = state

    @property
    def pending_state(self) -> str:
        """Gets the pending_state of this Sliver.


        :return: The pending_state of this Sliver.
        :rtype: str
        """
        return self._pending_state

    @pending_state.setter
    def pending_state(self, pending_state: str):
        """Sets the pending_state of this Sliver.


        :param pending_state: The pending_state of this Sliver.
        :type pending_state: str
        """

        self._pending_state = pending_state

    @property
    def join_state(self) -> str:
        """Gets the join_state of this Sliver.


        :return: The join_state of this Sliver.
        :rtype: str
        """
        return self._join_state

    @join_state.setter
    def join_state(self, join_state: str):
        """Sets the join_state of this Sliver.


        :param join_state: The join_state of this Sliver.
        :type join_state: str
        """

        self._join_state = join_state

    @property
    def graph_node_id(self) -> str:
        """Gets the graph_node_id of this Sliver.


        :return: The graph_node_id of this Sliver.
        :rtype: str
        """
        return self._graph_node_id

    @graph_node_id.setter
    def graph_node_id(self, graph_node_id: str):
        """Sets the graph_node_id of this Sliver.


        :param graph_node_id: The graph_node_id of this Sliver.
        :type graph_node_id: str
        """
        if graph_node_id is None:
            raise ValueError("Invalid value for `graph_node_id`, must not be `None`")  # noqa: E501

        self._graph_node_id = graph_node_id

    @property
    def slice_id(self) -> str:
        """Gets the slice_id of this Sliver.


        :return: The slice_id of this Sliver.
        :rtype: str
        """
        return self._slice_id

    @slice_id.setter
    def slice_id(self, slice_id: str):
        """Sets the slice_id of this Sliver.


        :param slice_id: The slice_id of this Sliver.
        :type slice_id: str
        """
        if slice_id is None:
            raise ValueError("Invalid value for `slice_id`, must not be `None`")  # noqa: E501

        self._slice_id = slice_id

    @property
    def sliver_id(self) -> str:
        """Gets the sliver_id of this Sliver.


        :return: The sliver_id of this Sliver.
        :rtype: str
        """
        return self._sliver_id

    @sliver_id.setter
    def sliver_id(self, sliver_id: str):
        """Sets the sliver_id of this Sliver.


        :param sliver_id: The sliver_id of this Sliver.
        :type sliver_id: str
        """
        if sliver_id is None:
            raise ValueError("Invalid value for `sliver_id`, must not be `None`")  # noqa: E501

        self._sliver_id = sliver_id

    @property
    def owner_user_id(self) -> str:
        """Gets the owner_user_id of this Sliver.


        :return: The owner_user_id of this Sliver.
        :rtype: str
        """
        return self._owner_user_id

    @owner_user_id.setter
    def owner_user_id(self, owner_user_id: str):
        """Sets the owner_user_id of this Sliver.


        :param owner_user_id: The owner_user_id of this Sliver.
        :type owner_user_id: str
        """

        self._owner_user_id = owner_user_id

    @property
    def owner_email(self) -> str:
        """Gets the owner_email of this Sliver.


        :return: The owner_email of this Sliver.
        :rtype: str
        """
        return self._owner_email

    @owner_email.setter
    def owner_email(self, owner_email: str):
        """Sets the owner_email of this Sliver.


        :param owner_email: The owner_email of this Sliver.
        :type owner_email: str
        """

        self._owner_email = owner_email



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import traceback
from typing import List

import threading

from fabric_mb.message_bus.consumer import AvroConsumerApi
from fabric_mb.message_bus.messages.abc_message_avro import AbcMessageAvro

from fabric_cf.actor.core.common.exceptions import KafkaServiceException


class MessageService(AvroConsumerApi):
    # uncomment for 1.7
    #def __init__(self, *, consumer_conf: dict, schema_registry_conf: dict, value_schema_location: str,
    #             topics: List[str], consumer_thread, batch_size: int = 5, logger: logging.Logger = None, sync: bool = False):
    #    super(MessageService, self).__init__(consumer_conf=consumer_conf, schema_registry_conf=schema_registry_conf,
    #                                         value_schema_location=value_schema_location, topics=topics,
    #                                         batch_size=batch_size, logger=logger)
    def __init__(self, *, consumer_conf: dict, key_schema_location, value_schema_location: str,
                 topics: List[str], consumer_thread, batch_size: int = 5, logger: logging.Logger = None,
                 poll_timeout: int = 250):
        super(MessageService, self).__init__(consumer_conf=consumer_conf, key_schema_location=key_schema_location,
                                             value_schema_location=value_schema_location, topics=topics,
                                             batch_size=batch_size, logger=logger, poll_timeout=poll_timeout)
        self.consumer_thread = consumer_thread
        self.thread_lock = threading.Lock()
        self.thread = None

    def start(self):
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise KafkaServiceException("This Message Service has already been started")

            self.thread = threading.Thread(target=self.consume, name=self.__class__.__name__,
                                           daemon=True)
            self.thread.start()
            self.logger.debug("Message service has been started")
        finally:
            self.thread_lock.release()

    def stop(self):
        self.shutdown()
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning("It seems that the Message Service thread is running. Interrupting it")
                try:
                    temp.join()
                except Exception as e:
                    self.logger.error("Could not join Message Service thread {}".format(e))
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def handle_message(self, message: AbcMessageAvro):
        try:
            self.consumer_thread.enqueue(message)
        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error(e)
            self.logger.error("Discarding the incoming message {}".format(message))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import timedelta, datetime, timezone
from http.client import BAD_REQUEST

from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_handler import OrchestratorHandler
from fabric_cf.orchestrator.swagger_server.models import Resource
from fabric_cf.orchestrator.swagger_server.models.resources import Resources  # noqa: E501
from fabric_cf.orchestrator.swagger_server import received_counter, success_counter, failure_counter
from fabric_cf.orchestrator.swagger_server.response.constants import GET_METHOD, RESOURCES_PATH, PORTAL_RESOURCES_PATH
from fabric_cf.orchestrator.swagger_server.response.utils import get_token, cors_error_response, cors_success_response


def portalresources_get(graph_format: str, level: int = 1, force_refresh: bool = False, start_date: str = None,
                        end_date: str = None, includes: str = None, excludes: str = None) -> Resources:  # noqa: E501
    """Retrieve a listing and description of available resources for portal

    Retrieve a listing and description of available resources for portal # noqa: E501

    :param graph_format: graph format
    :type graph_format: str
    :param level: Level of details
    :type level: int
    :param force_refresh: Force to retrieve current available resource information.
    :type force_refresh: bool
    :param start_date: starting date to check availability from
    :type start_date: str
    :param end_date: end date to check availability until
    :type end_date: str
    :param includes: comma separated lists of sites to include
    :type includes: str
    :param excludes: comma separated lists of sites to exclude
    :type excludes: str

    :rtype: Resources
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
    try:
        start = handler.validate_lease_time(lease_time=start_date)
        end = handler.validate_lease_time(lease_time=end_date)

        # Check if 'start' is defined but 'end' is not
        if start and not end:
            now = datetime.now(timezone.utc)

            # Check if the current time is within 10 minutes from 'start'
            if now - start < timedelta(minutes=10):
                # Reset start to None so as the cache is used
                start = None

        if start and end and (end - start) < timedelta(minutes=60):
            raise OrchestratorException(http_error_code=BAD_REQUEST,
                                        message="Time range should be at least 60 minutes long!")

        model = handler.list_resources(graph_format_str=graph_format, level=level, force_refresh=force_refresh,
                                       start=start, end=end, includes=includes, excludes=excludes, authorize=False)
        response = Resources()
        response.data = [Resource(model)]
        response.size = 1
        response.type = "resources"
        success_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, PORTAL_RESOURCES_PATH).inc()
        return cors_error_response(error=e)


def resources_get(level: int = 1, force_refresh: bool = False, start_date: str = None,
                  end_date: str = None, includes: str = None, excludes: str = None) -> Resources:  # noqa: E501
    """Retrieve a listing and description of available resources

    Retrieve a listing and description of available resources # noqa: E501

    :param level: Level of details
    :type level: int
    :param force_refresh: Force to retrieve current available resource information.
    :type force_refresh: bool
    :param start_date: starting date to check availability from
    :type start_date: str
    :param end_date: end date to check availability until
    :type end_date: str
    :param includes: comma separated lists of sites to include
    :type includes: str
    :param excludes: comma separated lists of sites to exclude
    :type excludes: str

    :rtype: Resources
    """
    handler = OrchestratorHandler()
    logger = handler.get_logger()
    received_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
    try:
        token = get_token()
        start = handler.validate_lease_time(lease_time=start_date)
        end = handler.validate_lease_time(lease_time=end_date)

        # Check if 'start' is defined but 'end' is not
        if start and not end:
            now = datetime.now(timezone.utc)

            # Check if the current time is within 10 minutes from 'start'
            if now - start < timedelta(minutes=10):
                # Reset start to None so as the cache is used
                start = None

        if start and end and (end - start) < timedelta(minutes=60):
            raise OrchestratorException(http_error_code=BAD_REQUEST,
                                        message="Time range should be at least 60 minutes long!")

        model = handler.list_resources(token=token, level=level, force_refresh=force_refresh,
                                       start=start, end=end, includes=includes, excludes=excludes)
        response = Resources()
        response.data = [Resource(model)]
        response.size = 1
        response.type = "resources"
        success_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
        return cors_success_response(response_body=response)
    except OrchestratorException as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
        return cors_error_response(error=e)
    except Exception as e:
        logger.exception(e)
        failure_counter.labels(GET_METHOD, RESOURCES_PATH).inc()
        return cors_error_response(error=e)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from collections import defaultdict
from datetime import datetime, timedelta

from fabric_cf.actor.core.time.actor_clock import ActorClock
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fim.slivers.base_sliver import BaseSliver
from fim.slivers.attached_components import ComponentSliver, ComponentType
from fim.slivers.capacities_labels import Capacities, FreeCapacity
from fim.slivers.network_node import NodeSliver


class TimeSlot:
    """Represents a time slot for resources, tracking capacities and components."""

    def __init__(self, end: datetime):
        """
        Initialize a TimeSlot instance with end time and empty resource capacities and components.

        :param end: The end datetime for this time slot.
        :type end: datetime
        """
        self.end = end
        self.capacities = Capacities()
        self.components = {}

    def __update_capacities(self, capacities: Capacities):
        """
        Update the capacities in this time slot.

        :param capacities: The capacities to add to this time slot.
        :type capacities: Capacities
        """
        self.capacities += capacities

    def __update_components(self, by_type: dict[ComponentType, list[ComponentSliver]]):
        """
        Update the components by type in this time slot.

        :param by_type: Dictionary with component types as keys and lists of ComponentSliver as values.
        :type by_type: dict[ComponentType, list[ComponentSliver]]
        """
        for comp_type, comps in by_type.items():
            if comp_type not in self.components:
                self.components[comp_type] = 0
            for c in comps:
                if c.get_capacities():
                    units = c.get_capacities().unit
                else:
                    units = c.get_capacity_allocations().unit
                self.components[comp_type] += units

    def add_sliver(self, sliver: BaseSliver):
        """
        Add sliver capacities and components to the current time slot.

        :param sliver: The sliver containing resource capacities and components to add.
        :type sliver: BaseSliver
        """
        if isinstance(sliver, NodeSliver):
            if sliver.capacity_allocations:
                self.__update_capacities(capacities=sliver.capacity_allocations)
            else:
                self.__update_capacities(capacities=sliver.capacities)

            if sliver.attached_components_info:
                self.__update_components(by_type=sliver.attached_components_info.by_type)

    def __str__(self):
        """
        Return a string representation of the capacities and components in this time slot.

        :return: String representation of capacities and components.
        :rtype: str
        """
        return f"Capacities: {self.capacities}, Components: {self.components}"


class ResourceTracker:
    """Tracks resource over time slots and checks availability of resources."""

    def __init__(self, cbm_node: NodeSliver):
        """
        Initialize ResourceTracker with total capacities and components from a CBM node.

        :param cbm_node: The CBM node from which to initialize capacities and components.
        :type cbm_node: NodeSliver
        """
        self.total_capacities = cbm_node.get_capacities()
        self.total_components = {}

        if cbm_node.attached_components_info:
            for comp_type, comps in cbm_node.attached_components_info.by_type.items():
                if comp_type not in self.total_components:
                    self.total_components[comp_type] = 0
                    for c in comps:
                        self.total_components[comp_type] += c.get_capacities().unit

        self.time_slots = defaultdict(TimeSlot)
        self.reservation_ids = set()

    def update(self, reservation: ReservationMng, start: datetime, end: datetime):
        """
        Update allocated resource information.

        :param reservation: Existing Reservation.
        :type reservation: ReservationMng
        :param start: Requested start datetime.
        :type start: datetime
        :param end: Requested end datetime.
        :type end: datetime
        """
        # Check if reservation has already been captured, if so skip it
        if reservation.get_reservation_id() in self.reservation_ids:
            return
        self.reservation_ids.add(reservation.get_reservation_id())

        start = start.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
        end = end.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
        sliver_end = ActorClock.from_milliseconds(milli_seconds=reservation.get_end())
        sliver_end = sliver_end.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)

        current_time = start
        while current_time < sliver_end and current_time < end:
            if current_time not in self.time_slots:
                self.time_slots[current_time] = TimeSlot(current_time)

            # Update the specific time slot with the sliver's resource usage
            self.time_slots[current_time].add_sliver(sliver=reservation.get_sliver())

            current_time += timedelta(hours=1)

    def __check_components(self, requested_sliver: NodeSliver,
                           allocated: dict[ComponentType, int]) -> bool:
        """
        Check if requested components can be fulfilled by available components.

        :param requested_sliver: The sliver with requested components.
        :type requested_sliver: NodeSliver
        :param allocated: Dictionary of available components by type.
        :type allocated: dict[ComponentType, int]
        :return: True if components can be fulfilled, False otherwise.
        :rtype: bool
        """
        if not requested_sliver.attached_components_info:
            return True
        for comp_type, comps in requested_sliver.attached_components_info.by_type.items():
            if comp_type not in self.total_components:
                return False
            elif len(comps) > (self.total_components[comp_type] - allocated.get(comp_type, 0)):
                return False
        return True

    def find_next_available(self, requested_sliver: NodeSliver, start: datetime, end: datetime,
                            duration: int) -> list[datetime]:
        """
        Find the next available time slot that can fulfill the requested sliver's capacities and components.

        :param requested_sliver: The sliver with requested capacities and components.
        :type requested_sliver: NodeSliver
        :param start: The start datetime to begin searching for availability.
        :type start: datetime
        :param end: The end datetime to stop searching.
        :type end: datetime
        :param duration: The duration in hours for which the resources are needed.
        :type duration: int
        :return: List of all possible next available time slot, or empty list if not found.
        :rtype: datetime
        """
        current_time = start.replace(minute=0, second=0, microsecond=0) + timedelta(hours=1)
        duration_timedelta = timedelta(hours=duration)
        ret_val = []

        while current_time + duration_timedelta <= end:
            available = True
            for i in range(duration):
                time_slot_time = current_time + timedelta(hours=i)
                if time_slot_time not in self.time_slots:
                    # If there's no entry for this time slot, assume full capacity available
                    continue

                time_slot = self.time_slots[time_slot_time]
                free_capacity = FreeCapacity(total=self.total_capacities, allocated=time_slot.capacities)

                # Check if accumulated capacities are negative (means not enough capacity)
                if free_capacity.free.negative_fields():
                    available = False
                    break

                diff = free_capacity.free - requested_sliver.capacities
                if diff.negative_fields():
                    available = False
                    break

                if not self.__check_components(requested_sliver=requested_sliver, allocated=time_slot.components):
                    available = False
                    break

            if available:
                ret_val.append(current_time)

            current_time += timedelta(hours=1)

        return ret_val



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING, List

from fim.graph.abc_property_graph import ABCPropertyGraph

from fabric_mb.message_bus.messages.add_slice_avro import AddSliceAvro
from fabric_mb.message_bus.messages.get_reservations_request_avro import GetReservationsRequestAvro
from fabric_mb.message_bus.messages.get_slices_request_avro import GetSlicesRequestAvro
from fabric_mb.message_bus.messages.reservation_mng import ReservationMng
from fabric_mb.message_bus.messages.slice_avro import SliceAvro

from fabric_cf.actor.core.apis.abc_mgmt_server_actor import ABCMgmtServerActor
from fabric_cf.actor.core.apis.abc_reservation_mixin import ReservationCategory
from fabric_cf.actor.core.common.constants import Constants
from fabric_cf.actor.core.common.exceptions import ManageException
from fabric_cf.actor.core.kernel.slice import SliceTypes
from fabric_cf.actor.core.manage.kafka.kafka_actor import KafkaActor
from fabric_cf.actor.core.manage.messages.client_mng import ClientMng
from fabric_cf.actor.security.auth_token import AuthToken


if TYPE_CHECKING:
    from fabric_cf.actor.core.util.id import ID


class KafkaServerActor(KafkaActor, ABCMgmtServerActor):
    def get_broker_reservations(self, *, id_token: str = None) -> List[ReservationMng]:
        request = GetReservationsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token)
        request.reservation_type = ReservationCategory.Broker.name
        status, response = self.send_request(request)

        if status.code == 0:
            return response.reservations
        return None

    def get_inventory_reservations(self, *, slice_id: ID = None, id_token: str = None) -> List[ReservationMng]:
        request = GetReservationsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token)
        request.reservation_type = ReservationCategory.Inventory.name
        status, response = self.send_request(request)

        if status.code == 0:
            return response.reservations
        return None

    def get_client_reservations(self, *, slice_id: ID = None, id_token: str = None) -> List[ReservationMng]:
        request = GetReservationsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token)
        request.reservation_type = ReservationCategory.Client.name
        status, response = self.send_request(request)

        if status.code == 0:
            return response.reservations
        return None

    def get_inventory_slices(self, *, id_token: str = None) -> List[SliceAvro]:
        request = GetReservationsRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token)
        request.type = SliceTypes.InventorySlice.name
        status, response = self.send_request(request)

        if status.code == 0:
            return response.slices
        return None

    def get_client_slices(self, *, id_token: str = None) -> List[SliceAvro]:
        request = GetSlicesRequestAvro()
        request = self.fill_request_by_id_message(request=request, id_token=id_token)
        request.type = SliceTypes.ClientSlice.name

        status, response = self.send_request(request)
        if response is not None:
            return response.slices

        return None

    def add_client_slice(self, *, slice_mng: SliceAvro) -> ID:
        ret_val = None
        request = AddSliceAvro()
        request.guid = str(self.management_id)
        request.auth = self.auth
        request.callback_topic = self.callback_topic
        request.message_id = str(ID())
        request.slice_obj = slice_mng

        status, response = self.send_request(request)

        if status.code == 0:
            ret_val = ID(uid=response.get_result())

        return ret_val

    def advertise_resources(self, *, delegation: ABCPropertyGraph, delegation_name: str, client: AuthToken) -> ID:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def get_clients(self, *, guid: ID = None, id_token: str = None) -> List[ClientMng]:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def register_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def update_client(self, *, client: ClientMng, kafka_topic: str) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)

    def unregister_client(self, *, guid: ID) -> bool:
        raise ManageException(Constants.NOT_IMPLEMENTED)



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric_cf.orchestrator.swagger_server.models.base_model_ import Model
from fabric_cf.orchestrator.swagger_server import util


class Status400BadRequestErrors(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, message: str='Bad Request', details: str=None, type: str='error', size: int=1, status: int=400):  # noqa: E501
        """Status400BadRequestErrors - a model defined in Swagger

        :param message: The message of this Status400BadRequestErrors.  # noqa: E501
        :type message: str
        :param details: The details of this Status400BadRequestErrors.  # noqa: E501
        :type details: str
        :param type: The type of this Status400BadRequestErrors.  # noqa: E501
        :type type: str
        :param size: The size of this Status400BadRequestErrors.  # noqa: E501
        :type size: int
        :param status: The status of this Status400BadRequestErrors.  # noqa: E501
        :type status: int
        """
        self.swagger_types = {
            'message': str,
            'details': str,
            'type': str,
            'size': int,
            'status': int
        }

        self.attribute_map = {
            'message': 'message',
            'details': 'details',
            'type': 'type',
            'size': 'size',
            'status': 'status'
        }
        self._message = message
        self._details = details
        self._type = type
        self._size = size
        self._status = status

    @classmethod
    def from_dict(cls, dikt) -> 'Status400BadRequestErrors':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The status_400_bad_request_errors of this Status400BadRequestErrors.  # noqa: E501
        :rtype: Status400BadRequestErrors
        """
        return util.deserialize_model(dikt, cls)

    @property
    def message(self) -> str:
        """Gets the message of this Status400BadRequestErrors.


        :return: The message of this Status400BadRequestErrors.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message: str):
        """Sets the message of this Status400BadRequestErrors.


        :param message: The message of this Status400BadRequestErrors.
        :type message: str
        """

        self._message = message

    @property
    def details(self) -> str:
        """Gets the details of this Status400BadRequestErrors.


        :return: The details of this Status400BadRequestErrors.
        :rtype: str
        """
        return self._details

    @details.setter
    def details(self, details: str):
        """Sets the details of this Status400BadRequestErrors.


        :param details: The details of this Status400BadRequestErrors.
        :type details: str
        """

        self._details = details

    @property
    def type(self) -> str:
        """Gets the type of this Status400BadRequestErrors.


        :return: The type of this Status400BadRequestErrors.
        :rtype: str
        """
        return self._type

    @type.setter
    def type(self, type: str):
        """Sets the type of this Status400BadRequestErrors.


        :param type: The type of this Status400BadRequestErrors.
        :type type: str
        """

        self._type = type

    @property
    def size(self) -> int:
        """Gets the size of this Status400BadRequestErrors.


        :return: The size of this Status400BadRequestErrors.
        :rtype: int
        """
        return self._size

    @size.setter
    def size(self, size: int):
        """Sets the size of this Status400BadRequestErrors.


        :param size: The size of this Status400BadRequestErrors.
        :type size: int
        """

        self._size = size

    @property
    def status(self) -> int:
        """Gets the status of this Status400BadRequestErrors.


        :return: The status of this Status400BadRequestErrors.
        :rtype: int
        """
        return self._status

    @status.setter
    def status(self, status: int):
        """Sets the status of this Status400BadRequestErrors.


        :param status: The status of this Status400BadRequestErrors.
        :type status: int
        """

        self._status = status



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import queue
import threading
import time
import traceback

from fabric_cf.actor.core.time.actor_clock import ActorClock

from fabric_cf.actor.core.kernel.reservation_states import ReservationStates
from fabric_cf.actor.core.util.id import ID
from fabric_cf.actor.core.util.iterable_queue import IterableQueue
from fabric_cf.orchestrator.core.exceptions import OrchestratorException
from fabric_cf.orchestrator.core.orchestrator_slice_wrapper import OrchestratorSliceWrapper


class AdvanceSchedulingThread:
    """
    This runs as a standalone thread started by Orchestrator and deals with determining the
    nearest start time for slices requested in future.
    The purpose of this thread is to help orchestrator respond back to the create
    without identifying the start time and waiting for the slivers to be demanded
    """

    def __init__(self, *, kernel):
        self.slice_queue = queue.Queue()
        self.slice_avail_condition = threading.Condition()
        self.thread_lock = threading.Lock()
        self.thread = None
        self.stopped = False
        from fabric_cf.actor.core.container.globals import GlobalsSingleton
        self.logger = GlobalsSingleton.get().get_logger()
        self.mgmt_actor = kernel.get_management_actor()
        self.kernel = kernel

    def queue_slice(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Queue a slice
        :param controller_slice:
        :return:
        """
        try:
            self.slice_queue.put_nowait(controller_slice)
            self.logger.debug(f"Added slice to slices queue {controller_slice.get_slice_id()}")
        except Exception as e:
            self.logger.error(f"Failed to queue slice: {controller_slice.get_slice_id()} e: {e}")

    def start(self):
        """
        Start thread
        :return:
        """
        try:
            self.thread_lock.acquire()
            if self.thread is not None:
                raise OrchestratorException(f"This {self.__class__.__name__} has already been started")

            self.thread = threading.Thread(target=self.run)
            self.thread.setName(self.__class__.__name__)
            self.thread.setDaemon(True)
            self.thread.start()

        finally:
            self.thread_lock.release()

    def stop(self):
        """
        Stop thread
        :return:
        """
        self.stopped = True
        try:
            self.thread_lock.acquire()
            temp = self.thread
            self.thread = None
            if temp is not None:
                self.logger.warning(f"It seems that the {self.__class__.__name__} is running. "
                                    f"Interrupting it")
                try:
                    # TODO find equivalent of interrupt
                    with self.slice_avail_condition:
                        self.slice_avail_condition.notify_all()
                    temp.join()
                except Exception as e:
                    self.logger.error(f"Could not join {self.__class__.__name__} thread {e}")
                finally:
                    self.thread_lock.release()
        finally:
            if self.thread_lock is not None and self.thread_lock.locked():
                self.thread_lock.release()

    def run(self):
        """
        Thread main loop
        :return:
        """
        self.logger.debug(f"{self.__class__.__name__} started")
        while True:
            slices = []
            if not self.stopped and self.slice_queue.empty():
                time.sleep(0.001)

            if self.stopped:
                self.logger.info(f"{self.__class__.__name__} exiting")
                return

            if not self.slice_queue.empty():
                try:
                    for s in IterableQueue(source_queue=self.slice_queue):
                        slices.append(s)
                except Exception as e:
                    self.logger.error(f"Error while adding slice to slice queue! e: {e}")
                    self.logger.error(traceback.format_exc())

            if len(slices) > 0:
                self.logger.debug(f"Processing {len(slices)} slices")
                for s in slices:
                    try:
                        # Process the Slice i.e. Determine the nearest start time
                        # If start time found, queue the slice on SliceDeferredThread
                        # Else move the slice to Close State with failure
                        self.process_slice(controller_slice=s)
                    except Exception as e:
                        self.logger.error(f"Error while processing slice {type(s)}, {e}")
                        self.logger.error(traceback.format_exc())

    def process_slice(self, *, controller_slice: OrchestratorSliceWrapper):
        """
        Determine nearest start time for a slice requested in future and
        add to deferred slice thread for further processing
        :param controller_slice:
        """
        computed_reservations = controller_slice.get_computed_reservations()

        try:
            controller_slice.lock()

            # Determine nearest start time in the time range requested
            # If not found, start time specified in the request is used as start resulting in slice failing
            # with insufficient resources error
            future_start, future_end = self.kernel.determine_future_lease_time(
                computed_reservations=computed_reservations,
                start=controller_slice.start, end=controller_slice.end,
                duration=controller_slice.lifetime)

            self.logger.debug(f"Slice: {controller_slice.slice_obj.slice_name}/{controller_slice.slice_obj.get_slice_id()}"
                              f" Start Time: {future_start} End: {future_end}")

            # Update slice start/end time
            controller_slice.slice_obj.set_lease_start(lease_start=future_start)
            controller_slice.slice_obj.set_lease_end(lease_end=future_end)
            self.logger.debug(f"Update Slice {controller_slice.slice_obj.slice_name}")
            self.mgmt_actor.update_slice(slice_obj=controller_slice.slice_obj)

            # Update the reservations start/end time
            for r in computed_reservations:
                r.set_start(value=ActorClock.to_milliseconds(when=future_start))
                r.set_end(value=ActorClock.to_milliseconds(when=future_end))

            self.add_and_demand_reservations(controller_slice=controller_slice)

        except Exception as e:
            self.logger.error(traceback.format_exc())
            self.logger.error("Unable to schedule a future slice: {}".format(e))
            self.add_and_demand_reservations(controller_slice=controller_slice)
        finally:
            controller_slice.unlock()

    def add_and_demand_reservations(self, controller_slice: OrchestratorSliceWrapper):
        try:
            # Add Reservations to relational database;
            controller_slice.add_reservations()

            # Queue the slice to be demanded on Slice Defer Thread
            self.kernel.get_defer_thread().queue_slice(controller_slice=controller_slice)
        except Exception as e:
            self.logger.error(f"SHOULD_NOT_HAPPEN: Queueing slice on Deferred Queue Failed: {controller_slice.slice_obj}")
            self.logger.error(f"Exception: {e}")
            self.logger.error(traceback.format_exc())
